/*
 *  network communications test using System V IPC's by Sami Kystil
 *
 *  history:
 * 
 *  v0.1:	initial revision
 *  v0.2:	added highscore system, changed a lot of stuff, fixed to
 * 		work with multiple instances on a single network
 *  v0.3:	fixed strange bugs with ncurses
 */

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/ipc.h>
#include <sys/msg.h>
#include <signal.h>
#include <unistd.h>
#include <time.h>

#include "vt100.h"
#include "ttt.h"

#define TTT_VERSION "0.3"

#define USERFILE "/home/hiteck/tttdata"
#define MAX_USERS 256

#define MAX_SEND_SIZE 80
#define PLAYER_1 42
#define PLAYER_2 43
#define M_SERVER 1
#define M_CLIENT 2

#define VT_BOLD "\x1B[1m"
#define VT_NORMAL "\x1B[0m"

struct mymsgbuf {
        long mtype;
        char mtext[MAX_SEND_SIZE];
};

typedef struct {
	char name[16];
	int win, lose, quit;
} user_data;

void send_message(int qid, struct mymsgbuf *qbuf, long type, char *text);
void read_message(int qid, struct mymsgbuf *qbuf, long type);
void remove_queue(int qid);
void change_queue_mode(int qid, char *mode);
void usage(void);

int  msgqueue_id, mode, state;

void signal_break(int a)
{
	fprintf(stderr, "Break, aborting...\n\r");
	remove_queue(msgqueue_id);
	exit(1);
}

void update_userdata(char *user, int win, int lose, int quit)
{
	FILE *handle = fopen(USERFILE, "r+");
		
	int i, found=0;
	long offset=0;

	user_data u;
	memset(&u, 0, sizeof(user_data));

        if (handle == NULL)
        {
                fprintf(stderr, "Unable to update data file.\n\r");
                return;
        }

	while (!feof(handle))
	{
		offset = ftell(handle);
		fread(&u, sizeof(user_data), 1, handle);
		if (strcmp(u.name, user)==0)
		{
			found=1;
			break;
		}
	}
	if (!found)
	{
		fseek(handle, 0, SEEK_END);
		offset = ftell(handle);
		memset(&u, 0, sizeof(user_data));
		u.win = 0;
		u.lose = 0;
		u.quit = 0;
	}

	fseek(handle, offset, SEEK_SET);
	strcpy(u.name, user);
	u.win += win;
	u.lose += lose;
	u.quit += quit;
	fwrite(&u, sizeof(user_data), 1, handle);	
	
	fclose(handle);
}

void view_top()
{
        FILE *handle = fopen(USERFILE, "r");

        int i=0, j, k;
        long offset=0, high=0;

        user_data *u = (user_data*)malloc(MAX_USERS*sizeof(user_data));
        memset(u, 0, sizeof(user_data)*MAX_USERS);

        if (handle == NULL)
        {
                fprintf(stderr, "Unable to open data file.\n\r");
                return;
        }

	
	for (;;)
        {
                fread(&u[i], sizeof(user_data), 1, handle);
		if (feof(handle)) break;
		if (u[i].win > high) high=u[i].win;
		i++;
	}
	printf("%d users.\n\r", i);
	k = 1;

	for (;high>=0; high--)
	for (j=0; j<i; j++)
		if (u[j].win==high)
		{
			if (k==1) printf(VT_BOLD);
			printf("%3d. %16s: won %4d, lost %4d, quit %4d\n\r", k++, u[j].name, u[j].win, u[j].lose, u[j].quit);
			if (k==2) printf(VT_NORMAL);
		}

	free(u);
	fclose(handle);
}


int main(int argc, char *argv[])
{
	int i, ox, oy;
        key_t key;
        struct mymsgbuf qbuf;
	char buffer[80];

	char lname[80], remote_name[80];
	cuserid(lname);

	signal(SIGINT, signal_break);

        printf("TTT v%s by Sami Kyostila 1999\n\r", TTT_VERSION);

	if (argc==2)
		if (tolower(argv[1][0])=='t') 
		{
			view_top();
			exit(1);
		}

        if(argc < 3)
        	usage();

	if (argc > 3) 
		keyset = argv[3];
	
	printf("Using keyset %s.\n\r", keyset);

	key = atoi(argv[2]);
	printf("Using key %ld.\n\r", key);

        if((msgqueue_id = msgget(key, IPC_CREAT|0666)) == -1) {
                perror("msgget");
                exit(1);
        }

	switch(tolower(argv[1][0]))
	{
		case 's':
		printf("%s%s%s is playing as 'server'.\n", VT_BOLD, lname, VT_NORMAL);
		mode = M_SERVER;
		fflush(stdout);
		break;

		case 'c':
		printf("%s%s%s is playing as 'client'.\n", VT_BOLD, lname, VT_NORMAL);
		mode = M_CLIENT;
		fflush(stdout);
		break;

		default:
		usage();
		break;
	}

	if (mode == M_SERVER)
	{
		printf("\nTelling client about ourselves...\n");
		send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_1, lname);

		printf("Waiting for client's name...\n");
                read_message(msgqueue_id, &qbuf, PLAYER_2);
                strcpy(remote_name, qbuf.mtext);
                printf("Playing against %s%s%s\n", VT_BOLD, remote_name, VT_NORMAL);
	} else {
		printf("\nWaiting for server's name...\n");
		read_message(msgqueue_id, &qbuf, PLAYER_1);
		strcpy(remote_name, qbuf.mtext);
		printf("Playing against %s%s%s\n", VT_BOLD, remote_name, VT_NORMAL);

 		printf("Telling server about ourselves...\n");
		send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_2, lname);
	}

        for (i=0; i<25; i++)
                printf("\n");
	fflush(stdout);

	initscr(); cbreak(); noecho(); nonl();
	refresh();

	cursor_set(56, 20);
	printf("IPCTTT by Sami Kystil\r\n");
	cursor_set(56, 21);
	printf("%c/%c/%c/%c/space/q\r\n", keyset[0], keyset[1], keyset[2], keyset[3]);
	
	cursor_set(1, 20);
	if (mode == M_SERVER)
	{
		printf("%sX: %s%s\n\r", VT_BOLD, lname, VT_NORMAL); 
		printf("O: %s\n\r", remote_name);
	} else {
                printf("X: %s\n\r", remote_name);
                printf("%sO: %s%s\n\r", VT_BOLD, lname, VT_NORMAL);
	}

	fflush(stdout);

	grid_init();
	grid_render();
	state = 0;
	fflush(stdout);
	fflush(stdin);

	for (;;)
	{
		if (mode == M_SERVER)
		{
			cursor_set(1, 19);
			printf("Your turn      ");
			state = place_mark("X");
			if (!state)
			{
				sprintf(buffer, "%c%c", (char)mark_x+1, (char)mark_y+1);
				send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_1, buffer);
			} else {
				buffer[0] = 0xff;
				send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_1, buffer);
                                cursor_set(1, 19);
                                printf("\x7User abort!         \n\r");
				update_userdata(lname, 0, 0, 1);
                                break;
			}

			if (check_for_win() == 'X')
			{
                        	cursor_set(1, 19);
                        	printf("\x7You won!        \n\r");
				update_userdata(lname, 1, 0, 0);
				break;
			}

                        cursor_set(1, 19);
                        printf("Opponent's turn");
                        fflush(stdout);
                        read_message(msgqueue_id, &qbuf, PLAYER_2);
                        ox = qbuf.mtext[0]-1;
                        oy = qbuf.mtext[1]-1;

                        if (ox==-2)
                        {
                                cursor_set(1, 19);
                                printf("\x7User abort!         \n\r");
                                break;
                        }


                        grid[ox][oy] = 'O';
                        grid_render();

                        if (check_for_win() == 'O')
                        {
                                cursor_set(1, 19);
                                printf("\x7You lost!         \n\r");
				update_userdata(lname, 0, 1, 0);
                                break;
                        }
		} else {
			cursor_set(1, 19);
			printf("Opponent's turn");
			fflush(stdout);
                	read_message(msgqueue_id, &qbuf, PLAYER_1);
			ox = qbuf.mtext[0]-1;
			oy = qbuf.mtext[1]-1;

			if (ox==-2)
			{
                                cursor_set(1, 19);
                                printf("\x7User abort!         \n\r");
                                break;
			}

			grid[ox][oy] = 'X';
			grid_render();
                        if (check_for_win() == 'X')
                        {
                                cursor_set(1, 19);
                                printf("\x7You lost!         \n\r");
				update_userdata(lname, 0, 1, 0);
                                break;
                        }

			
                        cursor_set(1, 19);
                        printf("Your turn      ");
                        state = place_mark("O");
                        if (!state)
                        {
                                sprintf(buffer, "%c%c", (char)mark_x+1, (char)mark_y+1);
                                send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_2, buffer);
                        } else {
                                buffer[0] = 0xff;
                                send_message(msgqueue_id, (struct mymsgbuf*)&qbuf, PLAYER_2, buffer);		
                                cursor_set(1, 19);
                                printf("\x7User abort!         \n\r");
				update_userdata(lname, 0, 0, 1);
                                break;
			}
			if (check_for_win() == 'O')
                        {
                                cursor_set(1, 19);
                                printf("\x7You won!        \n\r");
				update_userdata(lname, 1, 0, 0);
                                break;
                        }

		}
		if (state) break;
	}
	nocbreak(); echo();

	cursor_set(1,24);
	printf("Terminating connection...");
	fflush(stdout);
	mode = clock() + CLOCKS_PER_SEC;
	while (clock()<mode);
        remove_queue(msgqueue_id);

	printf("done.\n\r\n\r");

	endwin();
	return(0);
}

void send_message(int qid, struct mymsgbuf *qbuf, long type, char *text)
{
        /* Send a message to the queue */
        qbuf->mtype = type;
        strcpy(qbuf->mtext, text);

        if((msgsnd(qid, (struct msgbuf *)qbuf,
                strlen(qbuf->mtext)+1, 0)) ==-1)
        {
                perror("msgsnd");
                exit(1);
        }
}

void read_message(int qid, struct mymsgbuf *qbuf, long type)
{
        /* Read a message from the queue */
        qbuf->mtype = type;
        msgrcv(qid, (struct msgbuf *)qbuf, MAX_SEND_SIZE, type, 0);

/*        
        printf("Type: %ld Text: %s\n", qbuf->mtype, qbuf->mtext);
*/
}

void remove_queue(int qid)
{
        /* Remove the queue */
        msgctl(qid, IPC_RMID, 0);
}

void change_queue_mode(int qid, char *mode)
{
        struct msqid_ds myqueue_ds;

        /* Get current info */
        msgctl(qid, IPC_STAT, &myqueue_ds);

        /* Convert and load the mode */
        sscanf(mode, "%ho", &myqueue_ds.msg_perm.mode);

        /* Update the mode */
        msgctl(qid, IPC_SET, &myqueue_ds);
}

void usage(void)
{
        fprintf(stderr, "\n%sttt%s - Tick tack toe using System V IPC's by Sami Kyostila\n", VT_BOLD, VT_NORMAL);
        fprintf(stderr, "%susage:%s ttt [s(erver) | c(lient) | t(op)] [key] [keyset]\n", VT_BOLD, VT_NORMAL);

	fprintf(stderr, "\nFirst decide upon which player is the server and which is the client."
                        "\nThen pick a key number (just about any number), which both players must enter."
			"\nIf more than one ttt games are running on the network, the key must not conflict"
			"\nwith the keys of those games.\n"

			"\nThe keyset is a string of 4 characters that redefine the default movement keys"
			"\n(left, right, up, down), e.g. jlik would mean that the i-key moves the\ncursor left, j right, etc.\n"

			"\nIf you want to view the top score-files, use 'top' or 't' as the"
			"\nfirst parameter.\n");

	fprintf(stderr, "\n%skeys:%s\na/d/w/s   - left/right/up/down (default keys)\n", VT_BOLD, VT_NORMAL);
	fprintf(stderr, "space     - place mark\n");
	fprintf(stderr, "q         - abort\n");

	fprintf(stderr, "\nOh, and in case you didn't know, the objective of the game is to get 5 marks"
			"\nin a row before your opponent does.\n");
        exit(1);
}
