#include <stdio.h>
#include <stdlib.h>

#include "curses.h"

#define GRIDXSIZE 79
#define GRIDYSIZE 16
#define BLANKCHAR '.'

char grid[GRIDXSIZE][GRIDYSIZE];

char *keyset = "adws";

void grid_render()
{
	int x,y;
	char line[81];

        printf("\x1B[0m\x1B[1;1H");


	for (y=0; y<GRIDYSIZE; y++)
	{
		for (x=0; x<GRIDXSIZE; x++)
			line[x] = grid[x][y];

/*			printf("%c", grid[x][y]);*/
		line[80] = '\0';
		printf("%s\n\r", line);
	}
	fflush(stdout);
}

int mark_x=GRIDXSIZE/2, mark_y=GRIDYSIZE/2;

void draw_text(int x, int y, char *text)
{
	char buffer[256];
	sprintf(buffer, "\x1B[%d;%dH%s", y+1, x+1, text);
	printf(buffer);
	fflush(stdout);
}



void grid_draw_mark(int x, int y)
{
	int i;
	char text[8];

	printf("\x1B[0m");
	text[0] = grid[x][y];
	text[1] = '\0';

	draw_text(x, y, text);
	printf("\x1B[1D");
	fflush(stdout);
}

void grid_draw_inverse_mark(int x, int y)
{
	int i;
	char text[8];

	printf("\x1B[1m");
	text[0] = grid[x][y];
	text[1] = '\0';

	draw_text(x, y, text);
	printf("\x1B[1D");
	fflush(stdout);
}


int check_for_mark(int x, int y, char mark)
{
	if (x<0) return 0;
	if (y<0) return 0;
	if (x>GRIDXSIZE-1) return 0;
	if (y>GRIDYSIZE-1) return 0;

	if (grid[x][y]==mark) return 1;
	return 0;
}


char check_for_win()
{
	int x,y;
	char m;

	for (y=0; y<GRIDYSIZE; y++)
	for (x=0; x<GRIDXSIZE; x++)
	{
		m = grid[x][y];
		if (m != BLANKCHAR)
		{
			if (check_for_mark(x+1, y, m) &&
					check_for_mark(x+2, y, m) &&
					check_for_mark(x+3, y, m) &&
					check_for_mark(x+4, y, m)) return m;
	
			if (check_for_mark(x, y+1, m) &&
					check_for_mark(x, y+2, m) &&
					check_for_mark(x, y+3, m) &&
					check_for_mark(x, y+4, m)) return m;
	
	
			if (check_for_mark(x+1, y+1, m) &&
					check_for_mark(x+2, y+2, m) &&
					check_for_mark(x+3, y+3, m) &&
					check_for_mark(x+4, y+4, m)) return m;

                        if (check_for_mark(x-1, y+1, m) &&
                                        check_for_mark(x-2, y+2, m) &&
                                        check_for_mark(x-3, y+3, m) &&
                                        check_for_mark(x-4, y+4, m)) return m;
		}
	}

	return '\0';
}


void grid_init()
{
	memset(grid, BLANKCHAR, GRIDXSIZE*GRIDYSIZE);
}

int place_mark(char *mark)
{
	char key;

	grid_draw_inverse_mark(mark_x, mark_y);

	for (;;)
	{
		key = getch();
		grid_draw_mark(mark_x, mark_y);

		if (key == keyset[0]) mark_x--;
		if (key == keyset[1]) mark_x++;
		if (key == keyset[2]) mark_y--;
		if (key == keyset[3]) mark_y++;
		if (key == 'q') return 1;

		if (key == ' ')
		{
			if (grid[mark_x][mark_y]==BLANKCHAR)
			{
				grid[mark_x][mark_y] = mark[0];
				grid_draw_mark(mark_x, mark_y);
				return 0;
			}
		}

		if (mark_x<0) mark_x=0;
		if (mark_y<0) mark_y=0;

		if (mark_x>GRIDXSIZE-1) mark_x=GRIDXSIZE-1;
		if (mark_y>GRIDYSIZE-1) mark_y=GRIDYSIZE-1;

		grid_draw_inverse_mark(mark_x, mark_y);
	}
	return 0;
}
